import re
from dataclasses import dataclass


@dataclass(frozen=True)
class PageId:
    value_: str

    def __post_init__(self) -> None:
        # 型チェック
        if not isinstance(self.value_, str):
            msg = f"page_idは文字列である必要があります: {self.value_}"
            raise TypeError(msg)

        # UUID4の形式であることを確認する
        if not re.match(
            r"[0-9a-f]{8}-?[0-9a-f]{4}-?[0-9a-f]{4}-?[0-9a-f]{4}-?[0-9a-f]{12}",
            self.value_,
        ):
            msg = f"page_idの形式が不正です: {self.value_}"
            raise ValueError(msg)

    @staticmethod
    def dummy() -> "PageId":
        return PageId(value_="5c38fd30-714b-4ce2-bf2d-25407f3cfc16")

    @property
    def value(self) -> str:
        """UUID4の形式の文字列を返す"""
        # まずハイフンを削除してから、ハイフンをつけなおす
        value_ = self.value_.replace("-", "")
        return "-".join(
            [value_[:8], value_[8:12], value_[12:16], value_[16:20], value_[20:]]
        )



from dataclasses import dataclass

from .property import Property


@dataclass
class Checkbox(Property):
    checked: bool
    type: str = "checkbox"

    def __init__(
        self, name: str, checked: bool, id: str | None = None
    ) -> None:  # noqa: A002, FBT001
        self.name = name
        self.checked = checked or False
        self.id = id

    @staticmethod
    def of(name: str, param: dict) -> "Checkbox":
        return Checkbox(
            name=name,
            checked=param["checkbox"],
            id=param["id"],
        )

    @staticmethod
    def true(name: str) -> "Checkbox":
        return Checkbox(
            name=name,
            checked=True,
        )

    @staticmethod
    def false(name: str) -> "Checkbox":
        return Checkbox(
            name=name,
            checked=False,
        )

    def __dict__(self) -> dict:
        result = {
            "type": self.type,
            "checkbox": self.checked,
        }
        if self.id is not None:
            result["id"] = self.id
        return {self.name: result}

    def value_for_filter(self) -> str:
        raise NotImplementedError



from dataclasses import dataclass
from typing import Any

from .property import Property


@dataclass
class CreatedBy(Property):
    def __init__(
        self,
        name: str,
        created_by_param: dict,
        id: str | None = None,  # noqa: A002
    ) -> None:
        self.name = name
        self.created_by_param = created_by_param
        self.id = id

    @staticmethod
    def of(key: str, param: dict) -> "CreatedBy":
        return CreatedBy(id=param["id"], name=key, created_by_param=param["created_by"])

    def __dict__(self) -> dict[str, Any]:
        return {
            self.name: {
                "id": self.id,
                "type": self.type,
                "created_by": self.created_by_param,
            },
        }

    @property
    def type(self) -> str:
        return "created_by"  # NOTE: created_timeではなくdateにする

    def value_for_filter(self) -> str:
        raise NotImplementedError()



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass


@dataclass
class Block(metaclass=ABCMeta):
    id: str | None
    archived: bool | None
    has_children: bool | None
    created_time: str | None
    last_edited_time: str | None
    parent: dict[str, str] | None = None

    def to_dict(self, for_create: bool|None = None) -> dict:
        result = {
            "object": "block",
        }
        if not for_create:
            if self.id is not None:
                result["id"] = self.id
            if self.archived is not None:
                result["archived"] = self.archived
            if self.has_children is not None:
                result["has_children"] = self.has_children
            if self.created_time is not None:
                result["created_time"] = self.created_time
            if self.last_edited_time is not None:
                result["last_edited_time"] = self.last_edited_time
            if self.parent is not None:
                result["parent"] = self.parent
        result["type"] = self.type
        result[self.type] = self.to_dict_sub()
        return result

    @abstractmethod
    def to_dict_sub(self) -> dict:
        pass

    @property
    @abstractmethod
    def type(self) -> str:
        pass

    @abstractmethod
    def to_slack_text(self) -> str:
        pass



from dataclasses import dataclass
from datetime import date, datetime

from .property import Property
from ..datetime_utils import convert_to_date_or_datetime


@dataclass
class Date(Property):
    start: str | None = None
    end: str | None = None
    time_zone: str | None = None
    type: str = "date"

    DEFAULT_NAME = "日付"

    def __init__(
        self,
        name: str,
        id: str | None = None,  # noqa: A002
        start: str | None = None,
        end: str | None = None,
        time_zone: str | None = None,
    ) -> None:
        self.name = name
        self.id = id
        self.start = start
        self.end = end
        self.time_zone = time_zone

    @property
    def start_time(self) -> date | datetime | None:
        if self.start is None:
            return None
        return convert_to_date_or_datetime(self.start)

    @property
    def start_date(self) -> date | None:
        if self.start is None:
            return None
        return convert_to_date_or_datetime(self.start, cls=date)

    @property
    def start_datetime(self) -> datetime | None:
        if self.start is None:
            return None
        return convert_to_date_or_datetime(self.start, cls=datetime)

    @property
    def end_time(self) -> date | datetime | None:
        if self.end is None:
            return None
        return convert_to_date_or_datetime(self.end)

    @property
    def end_date(self) -> date | None:
        if self.end is None:
            return None
        return convert_to_date_or_datetime(self.end, cls=date)

    @property
    def end_datetime(self) -> datetime | None:
        if self.end is None:
            return None
        return convert_to_date_or_datetime(self.end, cls=datetime)

    @staticmethod
    def of(name: str = DEFAULT_NAME, param: dict | None = None) -> "Date":
        if param is None:
            param = {}
        if param["date"] is None:
            return Date(name=name, id=param["id"])
        return Date(
            name=name,
            id=param["id"],
            start=param["date"]["start"],
            end=param["date"]["end"],
            time_zone=param["date"]["time_zone"],
        )

    @staticmethod
    def from_start_date(
        name: str = DEFAULT_NAME, start_date: date | datetime | None = None
    ) -> "Date":
        return Date(
            name=name,
            start=start_date.isoformat() if start_date is not None else None,
        )

    @staticmethod
    def from_range(
        start: date | datetime, end: date | datetime, name: str = DEFAULT_NAME
    ) -> "Date":
        return Date(
            name=name,
            start=start.isoformat(),
            end=end.isoformat(),
        )

    def is_between(self, start: datetime, end: datetime) -> bool:
        start_datetime = self.start_datetime
        if start_datetime is None:
            return True
        if start.timestamp() > start_datetime.timestamp():
            return False
        if start_datetime.timestamp() > end.timestamp():  # noqa: SIM103
            return False
        return True

    @property
    def date(self) -> date:
        return convert_to_date_or_datetime(self.start, cls=date)

    def __dict__(self) -> dict:
        # 未指定の場合を考慮している
        _date = (
            {
                "start": self.start,
                "end": self.end,
                "time_zone": self.time_zone,
            }
            if self.start is not None
            else None
        )
        return {
            self.name: {
                "type": self.type,
                "date": _date,
            },
        }

    def value_for_filter(self) -> str:
        return self.start



import os
from datetime import datetime, timedelta
from logging import Logger, getLogger

from notion_client import Client
from notion_client.errors import APIResponseError, HTTPResponseError

from .base_operator import BaseOperator
from .base_page import BasePage
from .block import Block, BlockFactory
from .datetime_utils import JST
from .filter.builder import Builder
from .filter.condition.cond import Cond
from .filter.condition.prop import Prop
from .page.page_id import PageId
from .properties.cover import Cover
from .properties.icon import Icon
from .properties.multi_select import MultiSelect, MultiSelectElement, MultiSelectElements
from .properties.properties import Properties
from .properties.property import Property
from .properties.select import Select, Selects
from .property_translator import PropertyTranslator

NOTION_API_ERROR_BAD_GATEWAY = 502


class AppendBlockError(Exception):
    def __init__(self, block_id: str, blocks: list[dict], e: Exception) -> None:
        self.block_id = block_id
        self.blocks = blocks
        self.e = e
        super().__init__(f"block_id: {block_id}, blocks: {blocks}, error: {e}")


class NotionApiError(Exception):
    def __init__(
        self,
        page_id: str | None = None,
        database_id: str | None = None,
        e: APIResponseError | HTTPResponseError | None = None,
        properties: Properties | dict | None = None,
    ) -> None:
        self.database_id = database_id
        self.e = e
        self.properties = properties

        message = ""
        if e is not None:
            message += f", error: {e}"
        if page_id is not None:
            message += f"page_id: {page_id}"
        if database_id is not None:
            message += f"database_id: {database_id}"
        if properties is not None:
            properties_ = properties.__dict__() if isinstance(properties, Properties) else properties
            message += f", properties: {properties_}"
        super().__init__(message)


class Lotion:
    def __init__(self, client: Client, max_retry_count: int = 3, logger: Logger | None = None) -> None:
        self.client = client
        self.max_retry_count = max_retry_count
        self._logger = logger or getLogger(__name__)

    @staticmethod
    def get_instance(max_retry_count: int = 3, logger: Logger | None = None) -> "Lotion":
        client = Client(auth=os.getenv("NOTION_SECRET"))
        return Lotion(client, max_retry_count=max_retry_count, logger=logger)

    def retrieve_page(self, page_id: str) -> BasePage:
        """指定されたページを取得する"""
        page_entity = self.__retrieve_page(page_id=page_id)
        return self.__convert_page_model(page_entity=page_entity, include_children=True)

    def update_page(self, page_id: str, properties: list[Property] | None = None) -> dict:
        """指定されたページを更新する"""
        update_properties = Properties(values=properties or [])
        return self.__update(page_id=page_id, properties=update_properties)

    def retrieve_comments(self, page_id: str) -> list[dict]:
        """指定されたページのコメントを取得する"""
        comments = self.client.comments.list(
            block_id=page_id,
        )
        return comments["results"]

    def create_page_in_database(
        self,
        database_id: str,
        cover: Cover | None = None,
        properties: list[Property] | None = None,
        blocks: list[Block] | None = None,
    ) -> BasePage:
        """データベース上にページを新規作成する"""
        page = self.__create_page(
            database_id=database_id,
            cover=cover.__dict__() if cover is not None else None,
            properties=(Properties(values=properties).__dict__() if properties is not None else {}),
        )
        if blocks is not None:
            self.append_blocks(block_id=page["id"], blocks=blocks)
        return self.retrieve_page(page_id=page["id"])

    def retrieve_database(  # noqa: PLR0913
        self,
        database_id: str,
        filter_param: dict | None = None,
        include_children: bool | None = None,
    ) -> list[BasePage]:
        """指定されたデータベースのページを取得する"""
        results = self._database_query(database_id=database_id, filter_param=filter_param)
        pages: list[BasePage] = []
        for page_entity in results:
            page = self.__convert_page_model(
                page_entity=page_entity,
                include_children=include_children or False,
            )
            pages.append(page)
        return pages

    def find_page_by_title(
        self,
        database_id: str,
        title: str,
        title_key_name: str = "名前",
    ) -> BasePage | None:
        """タイトルだけをもとにデータベースのページを取得する"""
        filter_param = Builder.create().add(Prop.RICH_TEXT, title_key_name, Cond.EQUALS, title).build()
        results = self.retrieve_database(
            database_id=database_id,
            filter_param=filter_param,
        )
        if len(results) == 0:
            return None
        if len(results) > 1:
            warning_message = f"Found multiple pages with the same title: {title}"
            self._logger.warning(warning_message)
        return results[0]

    def find_page_by_unique_id(
        self,
        database_id: str,
        unique_id: int,
    ) -> BasePage | None:
        """UniqueIdをもとにデータベースのページを取得する"""
        unique_id_prop_name = None
        base_page = self._fetch_sample_page(database_id=database_id)
        for propety in base_page.properties.values:
            if propety.type == "unique_id":
                unique_id_prop_name = propety.name
                break

        if unique_id_prop_name is None:
            raise ValueError("unique_id property is not found")

        # filter_param = FilterBuilder.build_simple_equal_unique_id_condition(name=unique_id_prop_name, number=unique_id)
        filter_param = Builder.create().add(Prop.ID, unique_id_prop_name, Cond.EQUALS, unique_id).build()
        results = self.retrieve_database(
            database_id=database_id,
            filter_param=filter_param,
        )
        if len(results) == 0:
            return None
        return results[0]

    def _database_query(
        self,
        database_id: str,
        filter_param: dict | None = None,
        start_cursor: str | None = None,
    ) -> dict:
        if filter_param is None:
            return self._database_query_without_filter(database_id=database_id, start_cursor=start_cursor)
        results = []
        while True:
            data = self.__database_query(
                database_id=database_id,
                filter_param=filter_param,
                start_cursor=start_cursor,
            )
            results += data.get("results")
            if not data.get("has_more"):
                return results
            start_cursor = data.get("next_cursor")

    def _database_query_without_filter(self, database_id: str, start_cursor: str | None = None) -> dict:
        results = []
        while True:
            data = self.__database_query(
                database_id=database_id,
                start_cursor=start_cursor,
            )
            results += data.get("results")
            if not data.get("has_more"):
                return results
            start_cursor = data.get("next_cursor")

    def list_blocks(self, block_id: str) -> list[Block]:
        """指定されたブロックの子ブロックを取得する"""
        return self.__get_block_children(page_id=block_id)

    def append_block(self, block_id: str, block: Block) -> dict:
        """指定されたブロックに子ブロックを追加する"""
        return self.append_blocks(block_id=block_id, blocks=[block])

    def append_blocks(self, block_id: str, blocks: list[Block]) -> dict:
        """指定されたブロックに子ブロックを追加する"""
        return self.__append_block_children(
            block_id=block_id,
            children=[b.to_dict(for_create=True) for b in blocks],
        )

    def append_comment(self, page_id: str, text: str) -> dict:
        """指定されたページにコメントを追加する"""
        return self.client.comments.create(
            parent={"page_id": page_id},
            rich_text=[{"text": {"content": text}}],
        )

    def clear_page(self, page_id: str) -> None:
        """指定されたページのブロックを削除する"""
        blocks = self.list_blocks(block_id=page_id)
        for block in blocks:
            if block.id is None:
                raise ValueError(f"block_id is None: {block}")
            self.client.blocks.delete(block_id=block.id)

    def remove_page(self, page_id: str) -> None:
        """指定されたページを削除する"""
        self.__archive(page_id=page_id)

    def fetch_all_selects(self, database_id: str, name: str) -> Selects:
        """指定されたデータベースのセレクト一覧を取得する"""
        results = self.retrieve_database(database_id=database_id)
        selects = []
        for page in results:
            for prop in page.properties.values:
                if isinstance(prop, Select) and prop.name == name and not prop.is_empty():
                    selects.append(prop)
        return Selects(list(set(selects)))

    def fetch_select(self, database_id: str, name: str, status_name: str) -> Select:
        """
        指定されたデータベースのセレクトを取得する。
        ただし現在のデータベースで利用されていないセレクトを取得することはできない。
        """
        selects = self.fetch_all_selects(database_id=database_id, name=name)
        return selects.get(status_name)

    def fetch_all_multi_select_elements(self, database_id: str, name: str) -> MultiSelectElements:
        """指定されたデータベースのマルチセレクト一覧を取得する"""
        pages = self.retrieve_database(database_id=database_id)
        results: list[MultiSelectElement] = []
        for page in pages:
            for prop in page.properties.values:
                if isinstance(prop, MultiSelect) and prop.name == name:
                    results.extend(prop.values)
        return MultiSelectElements(list(set(results)))

    def fetch_multi_select(self, database_id: str, name: str, multi_select_names: list[str]) -> MultiSelect:
        """
        指定されたデータベースのマルチセレクトを取得する。
        ただし現在のデータベースで利用されていないマルチセレクトを取得することはできない。
        """
        all_elements = self.fetch_all_multi_select_elements(database_id=database_id, name=name)
        multi_element_list = all_elements.get(multi_select_names)
        return MultiSelect(name=name, values=multi_element_list)

    def __append_block_children(self, block_id: str, children: list[dict], retry_count: int = 0) -> dict:
        try:
            return self.client.blocks.children.append(block_id=block_id, children=children)
        except APIResponseError as e:
            if self.__is_able_retry(status=e.status, retry_count=retry_count):
                return self.__append_block_children(block_id=block_id, children=children, retry_count=retry_count + 1)
            raise NotionApiError(page_id=block_id, e=e) from e
        except HTTPResponseError as e:
            if self.__is_able_retry(status=e.status, retry_count=retry_count):
                return self.__append_block_children(block_id=block_id, children=children, retry_count=retry_count + 1)
            raise NotionApiError(page_id=block_id, e=e) from e
        except TypeError as e:
            raise AppendBlockError(block_id=block_id, blocks=children, e=e) from e

    def __convert_page_model(
        self,
        page_entity: dict,
        include_children: bool | None = None,
    ) -> BasePage:
        include_children = (
            include_children if include_children is not None else True
        )  # 未指定の場合はchildrenを取得する

        id_ = PageId(page_entity["id"])
        url = page_entity["url"]
        created_time = datetime.fromisoformat(page_entity["created_time"]) + timedelta(hours=9)
        last_edited_time = datetime.fromisoformat(page_entity["last_edited_time"]) + timedelta(hours=9)
        created_by = BaseOperator.of(page_entity["created_by"])
        last_edited_by = BaseOperator.of(page_entity["last_edited_by"])
        cover = Cover.of(page_entity["cover"]) if page_entity["cover"] is not None else None
        icon = Icon.of(page_entity["icon"]) if page_entity["icon"] is not None else None
        archived = page_entity["archived"]
        properties = PropertyTranslator.from_dict(page_entity["properties"])
        block_children = self.__get_block_children(page_id=id_.value) if include_children else []

        return BasePage(
            id_=id_,
            url=url,
            created_time=created_time.replace(tzinfo=JST),
            last_edited_time=last_edited_time.replace(tzinfo=JST),
            _created_by=created_by,
            _last_edited_by=last_edited_by,
            cover=cover,
            icon=icon,
            archived=archived,
            properties=properties,
            block_children=block_children,
        )

    def __retrieve_page(self, page_id: str, retry_count: int = 0) -> dict:
        try:
            return self.client.pages.retrieve(page_id=page_id)
        except APIResponseError as e:
            if self.__is_able_retry(status=e.status, retry_count=retry_count):
                return self.__retrieve_page(page_id=page_id, retry_count=retry_count + 1)
            raise NotionApiError(page_id=page_id, e=e) from e
        except HTTPResponseError as e:
            if self.__is_able_retry(status=e.status, retry_count=retry_count):
                return self.__retrieve_page(page_id=page_id, retry_count=retry_count + 1)
            raise NotionApiError(page_id=page_id, e=e) from e

    def __get_block_children(self, page_id: str) -> list[Block]:
        block_entities = self.__list_blocks(block_id=page_id)["results"]
        return [BlockFactory.create(b) for b in block_entities]

    def __list_blocks(self, block_id: str, retry_count: int = 0) -> dict:
        try:
            return self.client.blocks.children.list(block_id=block_id)
        except APIResponseError as e:
            if self.__is_able_retry(status=e.status, retry_count=retry_count):
                return self.__list_blocks(block_id=block_id, retry_count=retry_count + 1)
            raise NotionApiError(page_id=block_id, e=e) from e
        except HTTPResponseError as e:
            if self.__is_able_retry(status=e.status, retry_count=retry_count):
                return self.__list_blocks(block_id=block_id, retry_count=retry_count + 1)
            raise NotionApiError(page_id=block_id, e=e) from e

    def __archive(self, page_id: str, retry_count: int = 0) -> dict:
        try:
            return self.client.pages.update(
                page_id=page_id,
                archived=True,
            )
        except APIResponseError as e:
            if self.__is_able_retry(status=e.status, retry_count=retry_count):
                return self.__archive(page_id=page_id, retry_count=retry_count + 1)
            raise NotionApiError(page_id=page_id, e=e) from e
        except HTTPResponseError as e:
            if self.__is_able_retry(status=e.status, retry_count=retry_count):
                return self.__archive(page_id=page_id, retry_count=retry_count + 1)
            raise NotionApiError(page_id=page_id, e=e) from e

    def __update(self, page_id: str, properties: Properties, retry_count: int = 0) -> dict:
        try:
            return self.client.pages.update(
                page_id=page_id,
                properties=properties.exclude_for_update().__dict__(),
            )
        except APIResponseError as e:
            if self.__is_able_retry(status=e.status, retry_count=retry_count):
                return self.__update(page_id=page_id, properties=properties, retry_count=retry_count + 1)
            raise NotionApiError(page_id=page_id, e=e, properties=properties) from e
        except HTTPResponseError as e:
            if self.__is_able_retry(status=e.status, retry_count=retry_count):
                return self.__update(page_id=page_id, properties=properties, retry_count=retry_count + 1)
            raise NotionApiError(page_id=page_id, e=e, properties=properties) from e

    def __create_page(
        self,
        database_id: str,
        properties: dict,
        cover: dict | None = None,
        retry_count: int = 0,
    ) -> dict:
        try:
            return self.client.pages.create(
                parent={"type": "database_id", "database_id": database_id},
                cover=cover,
                properties=properties,
            )
        except APIResponseError as e:
            if self.__is_able_retry(status=e.status, retry_count=retry_count):
                self.__create_page(
                    database_id=database_id,
                    properties=properties,
                    cover=cover,
                    retry_count=retry_count + 1,
                )
            raise NotionApiError(database_id=database_id, e=e, properties=properties) from e
        except HTTPResponseError as e:
            if self.__is_able_retry(status=e.status, retry_count=retry_count):
                self.__create_page(
                    database_id=database_id,
                    properties=properties,
                    cover=cover,
                    retry_count=retry_count + 1,
                )
            raise NotionApiError(database_id=database_id, e=e, properties=properties) from e

    def _fetch_sample_page(self, database_id: str) -> BasePage:
        """指定されたデータベースのサンプルページを取得する"""
        data = self.__database_query(database_id=database_id, page_size=1)
        pages: list[dict] = data.get("results")
        if len(pages) == 0:
            raise ValueError(f"Database has no page. Please create any page. database_id: {database_id}")
        return self.__convert_page_model(page_entity=pages[0], include_children=False)

    def __database_query(
        self,
        database_id: str,
        start_cursor: str | None = None,
        filter_param: dict | None = None,
        page_size: int = 100,
        retry_count: int = 0,
    ) -> dict:
        try:
            if filter_param is None:
                return self.client.databases.query(
                    database_id=database_id,
                    start_cursor=start_cursor,
                    page_size=page_size,
                )
            return self.client.databases.query(
                database_id=database_id,
                start_cursor=start_cursor,
                filter=filter_param,
                page_size=page_size,
            )
        except APIResponseError as e:
            if self.__is_able_retry(status=e.status, retry_count=retry_count):
                return self.__database_query(
                    database_id=database_id,
                    start_cursor=start_cursor,
                    filter_param=filter_param,
                    retry_count=retry_count + 1,
                )
            raise NotionApiError(database_id=database_id, e=e) from e
        except HTTPResponseError as e:
            if self.__is_able_retry(status=e.status, retry_count=retry_count):
                return self.__database_query(
                    database_id=database_id,
                    start_cursor=start_cursor,
                    filter_param=filter_param,
                    retry_count=retry_count + 1,
                )
            raise NotionApiError(database_id=database_id, e=e) from e

    def __is_able_retry(self, status: int, retry_count: int) -> bool:
        return status == NOTION_API_ERROR_BAD_GATEWAY and retry_count < self.max_retry_count



from .block import Block


class ChildDatabase(Block):
    database_name: str

    def __init__(
        self,
        id: str,
        archived: bool,
        created_time: str,
        last_edited_time: str,
        has_children: bool,
        parent: dict,
        database_name: str,
    ):
        super().__init__(
            id, archived, created_time, last_edited_time, has_children, parent
        )
        self.database_name = database_name

    @staticmethod
    def of(block: dict) -> "ChildDatabase":
        database_name = block["child_database"]["title"]
        return ChildDatabase(
            id=block["id"],
            archived=block["archived"],
            created_time=block["created_time"],
            last_edited_time=block["last_edited_time"],
            has_children=block["has_children"],
            parent=block["parent"],
            database_name=database_name,
        )

    @property
    def type(self) -> str:
        return "child_database"

    def to_dict_sub(self) -> dict:
        raise NotImplementedError

    def to_slack_text(self) -> str:
        raise ""



from dataclasses import dataclass

from .property import Property


@dataclass
class Url(Property):
    url: str = ""

    def __init__(self, name: str, url: str = "", id: str | None = None):
        self.name = name
        self.url = url
        self.id = id

    @staticmethod
    def of(name: str, param: dict) -> "Url":
        url = param["url"] if param.get("url") else ""
        return Url(
            name=name,
            url=url,
            id=param["id"],
        )

    @staticmethod
    def from_url(url: str, name: str = "URL") -> "Url":
        return Url(
            name=name,
            url=url,
        )

    @staticmethod
    def empty(name: str = "URL") -> "Url":
        return Url(name=name)

    @property
    def type(self) -> str:
        return "url"

    def value_for_filter(self) -> str:
        return self.url

    def __dict__(self):
        result = {
            "type": self.type,
            "url": self.url if self.url != "" else None,
        }
        if self.id is not None:
            result["id"] = self.id
        return {
            self.name: result,
        }



from datetime import date, datetime, timedelta, timezone
from unittest import TestCase

from lotion.datetime_utils import convert_to_date_or_datetime

JST = timezone(timedelta(hours=+9), "JST")


class TestDatetime(TestCase):
    def test_日付だけ指定した場合でも日時で返す(self):
        # Given, When
        actual = convert_to_date_or_datetime("2021-01-01", cls=datetime)
        print(actual)

        # Then
        expected = datetime.fromisoformat("2021-01-01").replace(tzinfo=JST)
        print(expected)
        self.assertEqual(expected, actual)

    def test_0時0分の場合は日付として返却する(self):
        # Given, When
        actual = convert_to_date_or_datetime("2021-01-01 00:00:00+09:00")
        print(actual)

        # Then
        expected = date.fromisoformat("2021-01-01")
        print(expected)
        self.assertEqual(expected, actual)



from enum import Enum


class BlockType(Enum):
    VIDEO = "video"
    QUOTE = "quote"
    PARAGRAPH = "paragraph"
    HEADING_1 = "heading_1"
    TABLE = "table"
    CHILD_DATABASE = "child_database"
    COLUMN_LIST = "column_list"
    # 以下、Copilotによる自動生成
    HEADING_2 = "heading_2"
    HEADING_3 = "heading_3"
    BULLETED_LIST_ITEM = "bulleted_list_item"
    NUMBERED_LIST_ITEM = "numbered_list_item"
    TO_DO = "to_do"
    TOGGLE = "toggle"
    CHILD_PAGE = "child_page"
    EMBED = "embed"
    IMAGE = "image"
    COLUMN = "column"
    DIVIDER = "divider"
    TABLE_OF_CONTENTS = "table_of_contents"
    BREADCRUMB = "breadcrumb"
    FACTOID = "factoid"
    CALLOUT = "callout"
    CODE = "code"
    COLLECTION_VIEW = "collection_view"
    COLLECTION_VIEW_PAGE = "collection_view_page"
    EQUATION = "equation"
    FILE = "file"
    GALLERY = "gallery"
    LINK_TO_PAGE = "link_to_page"
    PAGE = "page"
    PDF = "pdf"
    SUB_SUB_HEADER = "sub_sub_header"
    SUB_HEADER = "sub_header"
    TWEET = "tweet"
    VIDEO_FILE = "video_file"
    AUDIO = "audio"
    BOOKMARK = "bookmark"
    CODE_BLOCK = "code_block"
    EMBED_BLOCK = "embed_block"
    FILE_BLOCK = "file_block"
    IMAGE_BLOCK = "image_block"
    PDF_BLOCK = "pdf_block"
    VIDEO_BLOCK = "video_block"
    AUDIO_BLOCK = "audio_block"
    BULLETED_LIST = "bulleted_list"



from dataclasses import dataclass
from .property import Property


@dataclass
class Select(Property):
    selected_name: str
    selected_id: str | None
    selected_color: str | None
    type: str = "select"

    def __init__(
        self,
        name: str,
        selected_name: str = "",
        selected_id: str | None = None,
        selected_color: str | None = None,
        id: str | None = None,
    ):
        self.name = name
        self.id = id
        self.selected_name = selected_name
        self.selected_id = selected_id
        self.selected_color = selected_color

    @staticmethod
    def of(name: str, param: dict) -> "Select":
        select = param["select"]
        if select is None:
            return Select(name=name)
        return Select(
            name=name,
            selected_id=select["id"],
            selected_name=select["name"],
            selected_color=select["color"],
            id=param["id"],
        )

    @staticmethod
    def empty(name: str) -> "Select":
        return Select(name=name)

    def is_empty(self) -> bool:
        return self.selected_name == ""

    def __dict__(self):
        if self.selected_id is None:
            return {
                self.name: {
                    "type": self.type,
                    "select": None,
                }
            }
        result = {
            "type": self.type,
            "select": {
                "id": self.selected_id,
                "name": self.selected_name,
                "color": self.selected_color,
            },
        }
        if self.id is not None:
            result["id"] = self.id
        return {self.name: result}

    def value_for_filter(self) -> str:
        return self.selected_name if self.selected_name is not None else ""

    # __hash__と__eq__を実装することで、リストやセットの中で比較が可能になる
    def __hash__(self):
        return hash(self.selected_id)

    def __eq__(self, other):
        return self.selected_id == other.selected_id


@dataclass(frozen=True)
class Selects:
    values: list[Select]

    def get(self, status_name: str) -> Select:
        for value in self.values:
            if value.selected_name == status_name:
                return value
        raise ValueError(f"Select not found: {status_name}")

    @property
    def size(self) -> int:
        return len(self.values)
